"""custom seabreeze exceptions

"""


class SeaBreezeError(Exception):
    """SeaBreezeError base class"""

    pass


class SeaBreezeNumFeaturesError(SeaBreezeError):
    pass


class SeaBreezeNotSupported(SeaBreezeError):
    pass



import os
import platform
import subprocess
import sys
import time
import warnings
from functools import wraps

import pytest


def _running_on_ci():
    """returns if we're currently running on a CI"""
    import os

    CI = os.environ.get("CI", "").lower()
    CONDA_BUILD = os.environ.get("CONDA_BUILD", "").lower()
    if CONDA_BUILD in {"1", "true"}:
        return True
    if CI in {"1", "true", "azure", "travis", "appveyor", "circleci"}:
        return True
    return False


RUNNING_ON_CI = _running_on_ci()
pytestmark = pytest.mark.skipif(
    RUNNING_ON_CI, reason="spectrometer tests are not run on ci"
)


def _aquire_connected_usb_spectrometers(timeout=10.0):
    """gathers the connected spectrometers for pytests"""
    # params and ids will be populated with the parametrization
    params = []
    ids = []
    supports = {}

    _skip = ([pytest.param("?", marks=pytest.mark.skip)], ["no-spectrometer"], 0, {})
    if RUNNING_ON_CI:
        # don't run on CI
        return _skip

    # execute seabreeze._cli.ls() in separate python process
    cmd = [sys.executable, "-c", "import seabreeze._cli as cli; cli.ls()"]
    p = subprocess.Popen(cmd, stdout=subprocess.PIPE)
    timeout_at = time.time() + timeout
    while p.poll() is None:
        if time.time() > timeout_at:
            p.kill()
            warnings.warn("killed scan for connected spectrometers!")
        time.sleep(0.5)

    # parse output to generate parametrization
    if p.returncode == 0:
        stdout, _ = p.communicate()
        for spectrometer in stdout.splitlines():
            spectrometer = spectrometer.decode("utf8")
            # parse line from command output
            model, serial_number, backends = spectrometer.split("\t")
            backends = backends.split(",")

            # set supports dict
            supports[serial_number] = backends

            # append the pytest.param
            params.append(serial_number)
            ids.append("{}:{}".format(model, serial_number))

    if not params:
        return _skip
    return params, ids, len(ids), supports


(
    _SPEC_PARAMS,
    _SPEC_IDS,
    _SPEC_NUM,
    _SPEC_SUPPORTS,
) = _aquire_connected_usb_spectrometers()


@pytest.fixture(scope="function", params=_SPEC_PARAMS, ids=_SPEC_IDS)
def serial_number(request):
    """yield serial numbers of connected spectrometers

    pytest.marks either cseabreeze or pyseabreeze dependent on support
    """
    yield request.param


@pytest.fixture(scope="class")
def backendlify(request):
    """setup the imports"""
    backend = request.param
    if backend == "any":
        pass

    elif backend in ("cseabreeze", "pyseabreeze"):
        import seabreeze

        seabreeze.use(backend)

    elif backend.startswith("pyseabreeze:"):
        _backend, _pyusb_backend = backend.split(":")
        import seabreeze

        seabreeze.use(_backend, _api_kwargs={"_pyusb_backend": _pyusb_backend})

    else:
        raise ValueError("unknown option '%s'" % backend)

    from seabreeze.backends import get_backend

    request.cls.backend = get_backend()
    yield


@pytest.fixture(scope="function")
def shutdown_api():
    try:
        yield
    finally:
        from seabreeze.spectrometers import list_devices

        api = getattr(list_devices, "_api", None)
        if api:
            api.shutdown()
            delattr(list_devices, "_api")
            time.sleep(0.1)


def skip_if_serial_unsupported_by_backend(arg_pos=0):
    def decorator(method):
        @wraps(method)
        def wrapper(self, *args, **kwargs):
            try:
                serial = kwargs["serial_number"]
            except KeyError:
                serial = args[arg_pos]
            # noinspection PyProtectedMember
            if self.backend._backend_ not in _SPEC_SUPPORTS[serial]:
                pytest.skip("spectrometer unsupported by backend")
            return method(self, *args, **kwargs)

        return wrapper

    return decorator


def skip_if_backend_and_os(backend, osname, pyusb_backend=None):
    def decorator(method):
        @wraps(method)
        def wrapper(self, *args, **kwargs):
            from seabreeze.backends import _SeaBreezeConfig as _Config

            # noinspection PyProtectedMember
            if (
                self.backend._backend_.startswith(backend)
                and platform.system() == osname
            ):
                requested = _Config["_api_kwargs"].get("_pyusb_backend", "any")

                if (
                    pyusb_backend is None
                    or requested == "any"
                    or pyusb_backend == requested
                ):
                    pytest.skip(
                        "skipping because {} on {}:{}".format(
                            backend, osname, requested
                        )
                    )
            return method(self, *args, **kwargs)

        return wrapper

    return decorator


# noinspection PyMethodMayBeStatic
@pytest.mark.usefixtures("backendlify")
@pytest.mark.usefixtures("shutdown_api")
class TestHardware(object):
    def test_cant_find_serial(self):
        from seabreeze.spectrometers import SeaBreezeError, Spectrometer

        with pytest.raises(SeaBreezeError):
            Spectrometer.from_serial_number("i-do-not-exist")

    @pytest.mark.skipif(_SPEC_NUM == 0, reason="no spectrometers connected")
    @skip_if_backend_and_os("pyseabreeze", "Darwin", "libusb1")
    def test_device_cleanup_on_exit(self):
        """test if opened devices cleanup correctly"""
        # noinspection PyProtectedMember
        backend_name = self.backend._backend_
        # noinspection PyProtectedMember
        from seabreeze.backends import _SeaBreezeConfig as _Config

        api_kwargs_str = ", ".join(
            "%s='%s'" % item for item in _Config["_api_kwargs"].items()
        )

        cmd = [
            sys.executable,
            "-c",
            "; ".join(
                [
                    "import seabreeze.%s as sb" % backend_name,
                    "d = sb.SeaBreezeAPI(%s).list_devices()[0]" % api_kwargs_str,
                    "d.open()",
                ]
            ),
        ]
        p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        stdout, stderr = p.communicate()
        assert p.returncode == 0, stderr

    @skip_if_serial_unsupported_by_backend()
    def test_read_model(self, serial_number):
        from seabreeze.spectrometers import Spectrometer

        spec = Spectrometer.from_serial_number(serial_number)
        model = spec.model
        assert len(model) > 0

    @skip_if_serial_unsupported_by_backend()
    def test_read_serial_number(self, serial_number):
        from seabreeze.spectrometers import Spectrometer

        spec = Spectrometer.from_serial_number(serial_number)
        serial = spec.serial_number
        assert len(serial) > 0

    @pytest.mark.xfail(reason="check if following tests work after crash")
    @pytest.mark.skipif(
        os.name == "nt", reason="FIXME: TEST BREAKS OTHER TESTS ON WINDOWS"
    )
    @skip_if_serial_unsupported_by_backend()
    def test_crash_may_not_influence_following_tests(self, serial_number):
        from seabreeze.spectrometers import Spectrometer

        _ = Spectrometer.from_serial_number(serial_number)
        raise Exception("crash on purpose")

    @skip_if_serial_unsupported_by_backend()
    def test_read_intensities(self, serial_number):
        from seabreeze.spectrometers import Spectrometer

        spec = Spectrometer.from_serial_number(serial_number)
        arr = spec.intensities()
        assert arr.size == spec.pixels

    @skip_if_serial_unsupported_by_backend()
    def test_correct_dark_pixels(self, serial_number):
        from seabreeze.spectrometers import SeaBreezeError, Spectrometer

        spec = Spectrometer.from_serial_number(serial_number)
        try:
            arr = spec.intensities(correct_dark_counts=True)
        except SeaBreezeError:
            pytest.skip("does not support dark counts")
        else:
            assert arr.size == spec.pixels
        finally:
            spec.close()

    @skip_if_serial_unsupported_by_backend()
    def test_read_wavelengths(self, serial_number):
        from seabreeze.spectrometers import Spectrometer

        spec = Spectrometer.from_serial_number(serial_number)
        arr = spec.wavelengths()
        assert arr.size == spec.pixels

    @skip_if_serial_unsupported_by_backend()
    def test_read_spectrum(self, serial_number):
        from seabreeze.spectrometers import Spectrometer

        spec = Spectrometer.from_serial_number(serial_number)
        arr = spec.spectrum()
        assert arr.shape == (2, spec.pixels)

    @skip_if_serial_unsupported_by_backend()
    def test_max_intensity(self, serial_number):
        from seabreeze.spectrometers import Spectrometer

        spec = Spectrometer.from_serial_number(serial_number)
        value = spec.max_intensity
        assert isinstance(value, float)
        assert 0 < value < 1e8

    @skip_if_serial_unsupported_by_backend()
    def test_integration_time_limits(self, serial_number):
        from seabreeze.spectrometers import Spectrometer

        spec = Spectrometer.from_serial_number(serial_number)
        low, high = spec.integration_time_micros_limits
        assert isinstance(low, int)
        assert isinstance(high, int)
        assert 0 < low < high < 2 ** 64

    @skip_if_serial_unsupported_by_backend()
    def test_integration_time(self, serial_number):
        from seabreeze.spectrometers import SeaBreezeError, Spectrometer

        spec = Spectrometer.from_serial_number(serial_number)

        with pytest.raises(SeaBreezeError):
            # fail because too low
            spec.integration_time_micros(0)

        with pytest.raises(SeaBreezeError):
            # fail because too large
            spec.integration_time_micros(2 ** 62)

        with pytest.raises(SeaBreezeError):
            # fail because too large long overflow
            spec.integration_time_micros(2 ** 64)

        spec.integration_time_micros(10000)
        spec.intensities()

    @skip_if_serial_unsupported_by_backend()
    def test_trigger_mode(self, serial_number):
        from seabreeze.spectrometers import Spectrometer

        spec = Spectrometer.from_serial_number(serial_number)
        spec.trigger_mode(0x00)  # <- normal mode

    @skip_if_serial_unsupported_by_backend()
    def test_trigger_mode_wrong(self, serial_number):
        from seabreeze.spectrometers import SeaBreezeError, Spectrometer

        spec = Spectrometer.from_serial_number(serial_number)
        with pytest.raises(SeaBreezeError):
            spec.trigger_mode(0xF0)  # <- should be unsupported for all specs

        spec.trigger_mode(0x00)  # <- normal mode
        spec.intensities()

    @pytest.mark.skipif(_SPEC_NUM == 0, reason="no spectrometers connected")
    def test_list_devices_dont_close_opened_devices(self):
        """test list_devices() interaction with already opened devices"""
        from seabreeze.spectrometers import list_devices

        devices = list_devices()
        num_devices = len(devices)

        dev = devices[0]
        dev.open()
        assert dev.is_open is True
        devices_2 = list_devices()
        assert len(devices_2) == num_devices
        # dev.is_open is still True here...
        assert dev.is_open is True
        # but if we delete the new reference created by the second call to list_devices:
        del devices_2
        assert dev.is_open is True



from seabreeze.pyseabreeze.features._base import SeaBreezeFeature


# Definition
# ==========
#
# TODO: This feature needs to be implemented for pyseabreeze
#
class SeaBreezeFastBufferFeature(SeaBreezeFeature):
    identifier = "fast_buffer"

    def get_buffering_enable(self):
        raise NotImplementedError("implement in derived class")

    def set_buffering_enable(self, is_enabled):
        raise NotImplementedError("implement in derived class")

    def get_consecutive_sample_count(self):
        raise NotImplementedError("implement in derived class")

    def set_consecutive_sample_count(self, consecutive_sample_count):
        raise NotImplementedError("implement in derived class")



from seabreeze.pyseabreeze.features._base import SeaBreezeFeature
from seabreeze.pyseabreeze.features.acquisitiondelay import (
    SeaBreezeAcquisitionDelayFeature,
)
from seabreeze.pyseabreeze.features.continuousstrobe import (
    SeaBreezeContinuousStrobeFeature,
)
from seabreeze.pyseabreeze.features.databuffer import SeaBreezeDataBufferFeature
from seabreeze.pyseabreeze.features.dhcpserver import SeaBreezeDHCPServerFeature
from seabreeze.pyseabreeze.features.eeprom import SeaBreezeEEPROMFeature
from seabreeze.pyseabreeze.features.ethernetconfiguration import (
    SeaBreezeEthernetConfigurationFeature,
)
from seabreeze.pyseabreeze.features.fastbuffer import SeaBreezeFastBufferFeature
from seabreeze.pyseabreeze.features.gpio import SeaBreezeGPIOFeature
from seabreeze.pyseabreeze.features.i2cmaster import SeaBreezeI2CMasterFeature
from seabreeze.pyseabreeze.features.introspection import SeaBreezeIntrospectionFeature
from seabreeze.pyseabreeze.features.ipv4 import SeaBreezeIPv4Feature
from seabreeze.pyseabreeze.features.irradcal import SeaBreezeIrradCalFeature
from seabreeze.pyseabreeze.features.lightsource import SeaBreezeLightSourceFeature
from seabreeze.pyseabreeze.features.multicast import SeaBreezeMulticastFeature
from seabreeze.pyseabreeze.features.networkconfiguration import (
    SeaBreezeNetworkConfigurationFeature,
)
from seabreeze.pyseabreeze.features.nonlinearity import (
    SeaBreezeNonlinearityCoefficientsFeature,
)
from seabreeze.pyseabreeze.features.opticalbench import SeaBreezeOpticalBenchFeature
from seabreeze.pyseabreeze.features.pixelbinning import SeaBreezePixelBinningFeature
from seabreeze.pyseabreeze.features.rawusb import SeaBreezeRawUSBBusAccessFeature
from seabreeze.pyseabreeze.features.revision import SeaBreezeRevisionFeature
from seabreeze.pyseabreeze.features.shutter import SeaBreezeShutterFeature
from seabreeze.pyseabreeze.features.spectrometer import SeaBreezeSpectrometerFeature
from seabreeze.pyseabreeze.features.spectrumprocessing import (
    SeaBreezeSpectrumProcessingFeature,
)
from seabreeze.pyseabreeze.features.straylightcoefficients import (
    SeaBreezeStrayLightCoefficientsFeature,
)
from seabreeze.pyseabreeze.features.strobelamp import SeaBreezeStrobeLampFeature
from seabreeze.pyseabreeze.features.temperature import SeaBreezeTemperatureFeature
from seabreeze.pyseabreeze.features.thermoelectric import SeaBreezeThermoElectricFeature
from seabreeze.pyseabreeze.features.wificonfiguration import (
    SeaBreezeWifiConfigurationFeature,
)



from seabreeze.pyseabreeze.exceptions import SeaBreezeError
from seabreeze.pyseabreeze.features._base import SeaBreezeFeature
from seabreeze.pyseabreeze.protocol import OOIProtocol


# Definition
# ==========
#
class SeaBreezeEEPROMFeature(SeaBreezeFeature):
    identifier = "eeprom"

    def eeprom_read_slot(self, slot_number, strip_zero_bytes=False):
        raise NotImplementedError("implement in derived class")


# OOI spectrometer implementation
# ===============================
#
class SeaBreezeEEPromFeatureOOI(SeaBreezeEEPROMFeature):
    _required_protocol_cls = OOIProtocol

    def eeprom_read_slot(self, slot_number, strip_zero_bytes=False):
        return self._func_eeprom_read_slot(
            self.protocol, slot_number, strip_zero_bytes=strip_zero_bytes
        )

    @staticmethod
    def _func_eeprom_read_slot(
        protocol, slot_number, strip_zero_bytes=False, raw=False
    ):
        protocol.send(0x05, slot_number)
        ret = protocol.receive(size=17, mode="low_speed")
        if ret[0] != 0x05 or ret[1] != int(slot_number) % 0xFF:
            raise SeaBreezeError('read_eeprom_slot_raw: wrong answer: "%s"' % ret)
        if raw:
            return ret
        try:
            data = ret[2 : ret[2:].index(0) + 2].tobytes().decode("utf-8")
        except AttributeError:  # Python27
            data = ret[2 : ret[2:].index(0) + 2].tostring().decode("utf-8")
        if not strip_zero_bytes:
            return data
        return data.rstrip("\x00")



from seabreeze.pyseabreeze.features._base import SeaBreezeFeature


# Definition
# ==========
#
# TODO: This feature needs to be implemented for pyseabreeze
#
class SeaBreezeTemperatureFeature(SeaBreezeFeature):
    identifier = "temperature"

    def count_temperatures(self):
        raise NotImplementedError("implement in derived class")

    def read_temperature(self, index):
        raise NotImplementedError("implement in derived class")

    def temperature_get_all(self):
        raise NotImplementedError("implement in derived class")



"""This is a wrapper for the c-implementation of the seabreeze-library

"""
# noinspection PyUnresolvedReferences,PyProtectedMember
from seabreeze.cseabreeze._wrapper import (
    SeaBreezeAcquisitionDelayFeature,
    SeaBreezeAPI,
    SeaBreezeContinuousStrobeFeature,
    SeaBreezeDataBufferFeature,
    SeaBreezeDevice,
    SeaBreezeDHCPServerFeature,
    SeaBreezeEEPROMFeature,
    SeaBreezeError,
    SeaBreezeEthernetConfigurationFeature,
    SeaBreezeFastBufferFeature,
    SeaBreezeFeature,
    SeaBreezeGPIOFeature,
    SeaBreezeI2CMasterFeature,
    SeaBreezeIntrospectionFeature,
    SeaBreezeIPv4Feature,
    SeaBreezeIrradCalFeature,
    SeaBreezeLightSourceFeature,
    SeaBreezeMulticastFeature,
    SeaBreezeNetworkConfigurationFeature,
    SeaBreezeNonlinearityCoefficientsFeature,
    SeaBreezeNumFeaturesError,
    SeaBreezeOpticalBenchFeature,
    SeaBreezePixelBinningFeature,
    SeaBreezeRawUSBBusAccessFeature,
    SeaBreezeRevisionFeature,
    SeaBreezeShutterFeature,
    SeaBreezeSpectrometerFeature,
    SeaBreezeSpectrumProcessingFeature,
    SeaBreezeStrayLightCoefficientsFeature,
    SeaBreezeStrobeLampFeature,
    SeaBreezeTemperatureFeature,
    SeaBreezeThermoElectricFeature,
    SeaBreezeWifiConfigurationFeature,
)

_backend_ = "cseabreeze"



"""seabreeze.compat compatibility layer with 0.6.x seabreeze

seabreeze 0.6.x

"""
import warnings


def _deprecation_warning(old_method, new_feature_method, version):
    msg = "{old_method} will be deprecated in version {ver}, use {feature} feature via f.{feature}.{method} instead"
    feature, method = new_feature_method.split(".")
    warnings.warn(
        msg.format(old_method=old_method, feature=feature, method=method, ver=version),
        PendingDeprecationWarning,
        stacklevel=2,
    )


class DeprecatedSpectrometerMixin(object):
    @property
    def f(self):
        """implemented in class using mixin"""
        return None

    def close(self):
        raise NotImplementedError("implemented in class using mixin")

    @property
    def integration_time_micros_limits(self):
        """implemented in class using mixin"""
        return None, None

    # Compatibility Layer Methods
    # ===========================
    #
    # will all go away in version 2
    #
    def boxcar_width(self, boxcar_width):
        _deprecation_warning(
            "boxcar_width", "spectrum_processing.set_boxcar_width", version="2"
        )
        self.f.spectrum_processing.set_boxcar_width(boxcar_width)

    def scans_to_average(self, scans_to_average):
        _deprecation_warning(
            "scans_to_average", "spectrum_processing.set_scans_to_average", version="2"
        )
        self.f.spectrum_processing.set_scans_to_average(self, scans_to_average)

    def get_boxcar_width(self):
        _deprecation_warning(
            "get_boxcar_width", "spectrum_processing.get_boxcar_width", version="2"
        )
        return self.f.spectrum_processing.get_boxcar_width()

    def get_scans_to_average(self):
        _deprecation_warning(
            "get_scans_to_average",
            "spectrum_processing.get_scans_to_average",
            version="2",
        )
        return self.f.spectrum_processing.get_scans_to_average()

    @property
    def minimum_integration_time_micros(self):
        return self.integration_time_micros_limits[0]

    @property
    def light_sources(self):
        _deprecation_warning("light_sources", "light_source.xxx", version="2")
        N = self.f.light_source.get_count()
        return tuple(DeprecatedLightSource(self.f.light_source, i) for i in range(N))

    def eeprom_read_slot(self, slot):
        _deprecation_warning("eeprom_read_slot", "eeprom.eeprom_read_slot", version="2")
        return self.f.eeprom.eeprom_read_slot(slot)

    def tec_set_enable(self, enable):
        _deprecation_warning(
            "tec_set_enable", "thermo_electric.enable_tec", version="2"
        )
        self.f.thermo_electric.enable_tec(enable)

    # noinspection PyPep8Naming
    def tec_set_temperature_C(self, set_point_C):
        _deprecation_warning(
            "tec_set_temperature_C",
            "thermo_electric.set_temperature_setpoint_degrees_celsius",
            version="2",
        )
        self.f.thermo_electric.set_temperature_setpoint_degrees_celsius(set_point_C)

    # noinspection PyPep8Naming
    def tec_get_temperature_C(self):
        _deprecation_warning(
            "tec_get_temperature_C",
            "thermo_electric.read_temperature_degrees_celsius",
            version="2",
        )
        return self.f.thermo_electric.read_temperature_degrees_celsius()

    def lamp_set_enable(self, enable):
        _deprecation_warning("lamp_set_enable", "strobe_lamp.enable_lamp", version="2")
        self.f.strobe_lamp.enable_lamp(enable)

    def shutter_set_open(self, state):
        _deprecation_warning(
            "shutter_set_open", "shutter.set_shutter_open", version="2"
        )
        self.f.shutter.set_shutter_open(state)

    def stray_light_coeffs(self):
        _deprecation_warning(
            "stray_light_coeffs",
            "stray_light_coefficients.get_stray_light_coefficients",
            version="2",
        )
        return self.f.stray_light_coefficients.get_stray_light_coefficients()

    def irrad_calibration(self):
        _deprecation_warning(
            "irrad_calibration", "irrad_cal.read_calibration", version="2"
        )
        return self.f.irrad_cal.read_calibration()

    def irrad_calibration_collection_area(self):
        _deprecation_warning(
            "irrad_calibration_collection_area",
            "irrad_cal.read_collection_area",
            version="2",
        )
        return self.f.irrad_cal.read_collection_area()

    def continuous_strobe_set_enable(self, enable):
        _deprecation_warning(
            "continuous_strobe_set_enable", "continuous_strobe.set_enable", version="2"
        )
        self.f.continuous_strobe.set_enable(enable)

    def continuous_strobe_set_period_micros(self, period_micros):
        _deprecation_warning(
            "continuous_strobe_set_period_micros",
            "continuous_strobe.set_period_micros",
            version="2",
        )
        self.f.continuous_strobe.set_period_micros(period_micros)

    def __enter__(self):
        warnings.warn(
            "context manager functionality will be a noop in version 2",
            PendingDeprecationWarning,
        )
        return self

    # noinspection PyShadowingBuiltins
    def __exit__(self, type, value, traceback):
        warnings.warn(
            "context manager functionality will be a noop in version 2",
            PendingDeprecationWarning,
        )
        self.close()


class DeprecatedLightSource(object):
    def __init__(self, spectrometer, index):
        self._id = "{}:{}".format(spectrometer.model, spectrometer.serial_number)
        self._f, self._idx = spectrometer.f.light_source, index
        self._has_enable = self._f.has_enable(self._idx)
        self._has_varint = self._f.has_variable_intensity(self._idx)

    def set_enable(self, enable):
        if self._has_enable:
            self._f.set_enable(self._idx, enable)
        else:
            from seabreeze.spectrometers import SeaBreezeError

            raise SeaBreezeError("Light source #%d can't be enabled or disabled.")

    def set_intensity(self, intensity):
        if self._has_varint:
            self._f.set_intensity(self._idx, intensity)
        else:
            from seabreeze.spectrometers import SeaBreezeError

            raise SeaBreezeError("Light source #%d intensity can't be set.")

    def get_intensity(self):
        from seabreeze.spectrometers import SeaBreezeError

        try:
            return self._f.get_intensity(self._idx)
        except SeaBreezeError:
            raise

    def __repr__(self):
        return "<DeprecatedLightSource #%d at %s>" % (self._idx, self._id)



from seabreeze.pyseabreeze.features._base import SeaBreezeFeature


# Definition
# ==========
#
# TODO: This feature needs to be implemented for pyseabreeze
#
class SeaBreezeDHCPServerFeature(SeaBreezeFeature):
    identifier = "dhcp_server"

    def get_address(self, interface_index):
        raise NotImplementedError("implement in derived class")

    def set_address(self, interface_index, server_address):
        raise NotImplementedError("implement in derived class")

    def get_enable_state(self, interface_index):
        raise NotImplementedError("implement in derived class")

    def set_enable_state(self, interface_index, enable_state):
        raise NotImplementedError("implement in derived class")



from seabreeze.pyseabreeze.features._base import SeaBreezeFeature


# Definition
# ==========
#
# TODO: This feature needs to be implemented for pyseabreeze
#
class SeaBreezeIntrospectionFeature(SeaBreezeFeature):
    identifier = "introspection"

    def number_of_pixels(self):
        raise NotImplementedError("implement in derived class")

    def get_active_pixel_ranges(self):
        raise NotImplementedError("implement in derived class")

    def get_optical_dark_pixel_ranges(self):
        raise NotImplementedError("implement in derived class")

    def get_electric_dark_pixel_ranges(self):
        raise NotImplementedError("implement in derived class")



from seabreeze.pyseabreeze.features._base import SeaBreezeFeature
from seabreeze.pyseabreeze.protocol import ProtocolInterface
from seabreeze.pyseabreeze.transport import USBTransport


class SeaBreezeRawUSBBusAccessFeature(SeaBreezeFeature):
    r"""
    Reading and writing raw usb

    Example usage

    .. code-block:: python

     >>> import struct  # needed for packing binary data into bytestrings
     >>> from seabreeze.spectrometers import Spectrometer
     >>> spec = Spectrometer.from_first_available()
     # features need to be accessed on the spectrometer instance via spec.f or spec.features
     # the data you provide needs to packed into a bytestring via struct.pack
     >>> spec.f.raw_usb_bus_access.raw_usb_write(data=struct.pack('<B', 0xFE), endpoint='primary_out')

     # when reading via the raw_usb feature you can easily block the spectrometer.
     # so make sure to only read when you expect new data (after you sent a command)
     # and only read as many bytes as the command returns
     >>> output = spec.f.raw_usb_bus_access.raw_usb_read(endpoint='primary_in', buffer_length=16)

     >>> print(output)
     b'\x00\x08p\x17\x00\x00\x00\x00\x00\x08\x00\x00\x00\x00\x80U'
     # extract by providing the data format from the developer datasheet and `output` to struct.unpack

    """
    identifier = "raw_usb_bus_access"
    _required_protocol_cls = ProtocolInterface

    @classmethod
    def supports_protocol(cls, protocol):
        return isinstance(protocol.transport, USBTransport)

    def raw_usb_write(self, data, endpoint):
        if endpoint not in {"primary_out", "secondary_out"}:
            raise ValueError("endpoint has to be one of 'primary_out', 'secondary_out'")
        if endpoint == "secondary_out":
            pass
        self.protocol.transport.write(data)

    def raw_usb_read(self, endpoint, buffer_length=None):
        if endpoint == "primary_in":
            return self.protocol.transport.read(size=buffer_length, mode="low_speed")
        elif endpoint == "secondary_in":
            return self.protocol.transport.read(size=buffer_length, mode="high_speed")
        elif endpoint == "secondary_in2":
            return self.protocol.transport.read(
                size=buffer_length, mode="high_speed_alt"
            )
        else:
            raise ValueError(
                "endpoint has to be one of 'primary_in', 'secondary_in', 'secondary_in2'"
            )



# test both backends
import pytest


def test_seabreeze_installed():
    """if this test fails, you forgot to install seabreeze in your environment"""
    # noinspection PyUnresolvedReferences
    import seabreeze


def test_seabreeze_wrong_backend_requested():
    """requests a nonexistent backend"""
    import seabreeze

    with pytest.raises(ValueError):
        seabreeze.use("i-do-not-exist")


@pytest.mark.parametrize("backend_name", ["cseabreeze", "pyseabreeze"])
def test_seabreeze_any_backend_available(backend_name):
    """requests one backend and allows fallback"""
    import seabreeze

    seabreeze.use(backend_name, force=False)
    from seabreeze.backends import get_backend

    get_backend()


def test_seabreeze_cseabreeze_backend_available():
    """tests if cseabreeze was compiled on install"""
    try:
        import seabreeze.cseabreeze
    except ImportError:
        pytest.skip("can't import cseabreeze")


def test_seabreeze_pyseabreeze_backend_available():
    """basically just tests if pyusb is installed"""
    try:
        import seabreeze.pyseabreeze
    except ImportError:
        pytest.skip("can't import pyseabreeze")


def test_seabreeze_cseabreeze_api_init(cseabreeze):
    """check if SeaBreezeAPI can be instantiated"""
    cseabreeze.SeaBreezeAPI()


@pytest.mark.usefixtures("mock_pyusb_core_find")
def test_seabreeze_pyseabreeze_api_init(pyseabreeze, pyseabreeze_pyusb_backend):
    """check if SeaBreezeAPI can be instantiated"""
    try:
        pyseabreeze.SeaBreezeAPI(_pyusb_backend=pyseabreeze_pyusb_backend)
    except RuntimeError:
        if pyseabreeze_pyusb_backend is None:
            raise  # fail if no backend library is available
        else:
            pytest.skip(
                "pyusb_backend {} not available".format(pyseabreeze_pyusb_backend)
            )


def _get_class_public_interface_dict(backend):
    """return a dictionary with a set of all public functions for each feature"""
    base_class = backend.SeaBreezeFeature
    feature_classes = base_class.__subclasses__()
    interface_dict = {}
    for fcls in feature_classes:
        name = fcls.__name__
        attrs = set(attr for attr in dir(fcls) if not attr.startswith("_"))
        interface_dict[name] = attrs
    return interface_dict


def test_seabreeze_compare_backend_feature_interfaces(cseabreeze, pyseabreeze):
    """test if the public backend feature class interfaces are identical"""
    c_feature_interface = _get_class_public_interface_dict(cseabreeze)
    py_feature_interface = _get_class_public_interface_dict(pyseabreeze)

    # test if the interfaces contain something
    assert c_feature_interface and py_feature_interface

    # test if all feature interfaces are non empty
    assert all(c_feature_interface.values()) and all(py_feature_interface.values())

    # test if there's a difference between cseabreeze and pyseabreeze features
    assert set(c_feature_interface) == set(py_feature_interface)

    CSEABREEZE_CUSTOM = {"feature_id"}
    PYSEABREEZE_CUSTOM = {"supports_protocol", "specialize"}

    for feature in c_feature_interface:
        c_attrs = c_feature_interface[feature] - CSEABREEZE_CUSTOM
        py_attrs = py_feature_interface[feature] - PYSEABREEZE_CUSTOM
        # for each feature test if there's a difference between cseabreeze and pyseabreeze
        assert c_attrs == py_attrs, "feature {} differs in attrs".format(feature)



import mock
import pytest


def pytest_addoption(parser):
    pyusb_choices = [
        "any",
        "all",
        "cseabreeze",
        "pyseabreeze",
        "pyseabreeze:libusb0",
        "pyseabreeze:libusb1",
        "pyseabreeze:openusb",
    ]
    parser.addoption(
        "--seabreeze-backend",
        default="any",
        choices=pyusb_choices,
        help="options: %s" % ", ".join(pyusb_choices),
    )


def pytest_generate_tests(metafunc):
    if "backendlify" in metafunc.fixturenames:
        backend = metafunc.config.getoption("--seabreeze-backend")
        if backend == "all":
            argvalues = [
                "cseabreeze",
                "pyseabreeze:libusb0",
                "pyseabreeze:libusb1",
                "pyseabreeze:openusb",
            ]
        else:
            argvalues = [backend]
        metafunc.parametrize(
            "backendlify", argvalues=argvalues, ids="backend({})".format, indirect=True,
        )


@pytest.fixture(scope="function")
def mock_pyusb_core_find():
    """mock usb.core.find

    when running on a ci, the pyusb can crash if access
    to the usbfs is denied: https://github.com/pyusb/pyusb/issues/151

    """
    try:
        with mock.patch("usb.core.find", return_value=[]):
            yield
    except ImportError:
        pytest.mark.skip(reason="no pyusb installed")


@pytest.fixture(scope="module")
def cseabreeze():
    """try importing the cseabreeze backend"""
    yield pytest.importorskip("seabreeze.cseabreeze")


@pytest.fixture(scope="module")
def pyseabreeze():
    """try importing the pyseabreeze backend"""
    yield pytest.importorskip("seabreeze.pyseabreeze")


@pytest.fixture(
    scope="function",
    params=[None, "openusb", "libusb0", "libusb1"],
    ids=["any", "openusb", "libusb0", "libusb1"],
)
def pyseabreeze_pyusb_backend(request):
    yield request.param



"""device classes for pyseabreeze

Author: Andreas Poehlmann
"""
import enum
import itertools
from collections import defaultdict

from future.utils import with_metaclass

from seabreeze.pyseabreeze import features as sbf
from seabreeze.pyseabreeze.exceptions import SeaBreezeError
from seabreeze.pyseabreeze.features import SeaBreezeFeature
from seabreeze.pyseabreeze.protocol import OBPProtocol, OOIProtocol
from seabreeze.pyseabreeze.transport import TransportInterface, USBTransport

# class registry for all spectrometer models
_model_class_registry = {}


class _SeaBreezeDeviceMeta(type):
    """metaclass for pyseabreeze devices"""

    def __new__(mcs, name, bases, attr_dict):
        # This part of the metaclass magic is very opaque. It could be avoided by moving all
        # of this logic to custom __init__ methods of the Spectrometer classes, or to factory
        # functions. But I am using python-seabreeze as a playground to experiment with
        # different ways of implementing extendable interfaces, so please forgive these design
        # decisions ^^"
        #
        # I also like the concise way you can define the spectrometer functionality this way.
        # It allows you to get a quick overview of the functionality implemented in pyseabreeze,
        # without having to read through a lot of code. And I think that it's easy to understand
        # what a spectrometer can do, by just looking at the spectrometer class definition, even
        # if you don't understand the details of the implementation.
        #
        # Because it's quite opaque the code below tries hard to enforce strict typing in all
        # defined spectrometer classes to minimize the amount of errors you can make when adding
        # a new spectrometer to pyseabreeze.
        #
        if name != "SeaBreezeDevice":
            # This runs for all subclasses of SeaBreezeDevice, so for all defined spectrometers
            # What might be unintuitive to the user is, that all defined attributes in the
            # spectrometer classes are only used to configure the subclass and are NOT directly
            # available in the instances later. (look at how attr_dict is modified below).
            #
            if "model_name" not in attr_dict:
                raise AttributeError(
                    "'model_name' not provided for class '{}'".format(name)
                )
            model_name = attr_dict.pop("model_name")
            if not isinstance(model_name, str):
                raise TypeError("{}.model_name not a str".format(name))

            # gather the transport classes defined on the class
            transport_classes = mcs._extract_transform_classes(
                model_name, class_name=name, attr_dict=attr_dict
            )
            # gather the feature classes defined on the class
            feature_classes = mcs._extract_feature_classes(
                model_name, class_name=name, attr_dict=attr_dict
            )

            if any(not attr.startswith("_") for attr in attr_dict):
                raise ValueError(
                    "can't define extra attrs on spectrometer classes: {}".format(
                        ", ".join(
                            attr for attr in attr_dict if not attr.startswith("_")
                        )
                    )
                )

            attr_dict = {
                "_model_name": model_name,
                "_transport_classes": transport_classes,
                "_feature_classes": feature_classes,
            }

        return super(_SeaBreezeDeviceMeta, mcs).__new__(mcs, name, bases, attr_dict)

    def __init__(cls, name, bases, attr_dict):
        if name != "SeaBreezeDevice":
            # > model name
            model_name = getattr(cls, "_model_name")
            assert isinstance(model_name, str), "model name not a str"
            _model_class_registry[model_name] = cls

        super(_SeaBreezeDeviceMeta, cls).__init__(name, bases, attr_dict)

    @staticmethod
    def _extract_transform_classes(model_name, class_name, attr_dict):

        visited_attrs = set()
        transport_classes = []
        try:
            supported_transport_classes = attr_dict.pop("transport")
        except KeyError:
            raise AttributeError("{}.transport not provided")
        if (
            not isinstance(supported_transport_classes, tuple)
            or not supported_transport_classes
        ):
            raise TypeError("{}.transport not a tuple of len > 0")

        for idx, transport_cls in enumerate(supported_transport_classes):
            # for each supported transport of the spectrometer, gather the configuration from
            # the spectrometer class and specialize the transport_cls with the provided settings.
            #
            if not issubclass(transport_cls, TransportInterface):
                raise TypeError(
                    "{}.transport[{:d}] '{}' does not derive from TransportInterface".format(
                        class_name, idx, transport_cls.__name__
                    )
                )
            # noinspection PyProtectedMember
            kwargs = transport_cls._required_init_kwargs
            transport_init_kwargs = {}
            for kw in kwargs:
                if kw not in attr_dict:
                    raise AttributeError(
                        "{}.{} not provided for class but '{}' requires it.".format(
                            class_name, kw, transport_cls.__name__
                        )
                    )
                transport_init_kwargs[kw] = attr_dict[kw]
            visited_attrs.update(kwargs)
            # specialize the transport class with the spectrometer's custom config
            specialized_transport_cls = transport_cls.specialize(
                model_name, **transport_init_kwargs
            )
            transport_classes.append(specialized_transport_cls)

        for attr in visited_attrs:
            del attr_dict[attr]

        return tuple(transport_classes)

    @staticmethod
    def _extract_feature_classes(model_name, class_name, attr_dict):

        visited_attrs = set()
        feature_classes = defaultdict(list)
        try:
            supported_feature_classes = attr_dict.pop("feature_classes")
        except KeyError:
            raise AttributeError("{}.feature_classes not provided")
        if (
            not isinstance(supported_feature_classes, tuple)
            or not supported_feature_classes
        ):
            raise TypeError("{}.feature_classes not a tuple of len > 0")
        for idx, feature_cls in enumerate(supported_feature_classes):
            # for each supported feature of the spectrometer, gather the configuration
            # from the spectrometer class and subclass the feature_cls with the provided
            # settings. Also check if requirements are fulfilled, i.e. if the feature depends
            # on other features or on specific protocols
            #
            if not issubclass(feature_cls, SeaBreezeFeature):
                raise TypeError(
                    "{}.feature_classes[{:d}] '{}' does not derive from SeaBreezeFeature".format(
                        model_name, idx, feature_cls.__name__
                    )
                )
            # noinspection PyProtectedMember
            required = set(feature_cls._required_features)
            if not required.issubset(feature_classes):
                raise KeyError(
                    "{}.feature_classes[{:d}] '{}' requires '{}'. To fix, re-order or add.".format(
                        model_name,
                        idx,
                        feature_cls.__name__,
                        ", ".join(required - set(feature_classes)),
                    )
                )
            # noinspection PyProtectedMember
            kwargs = feature_cls._required_kwargs
            feature_attrs = {}
            for kw in kwargs:
                if kw not in attr_dict:
                    raise AttributeError(
                        "{}.{} not provided for class but '{}' requires it.".format(
                            class_name, kw, feature_cls.__name__
                        )
                    )
                feature_attrs[kw] = attr_dict[kw]
            visited_attrs.update(kwargs)
            # specialize the feature class with the spectrometer's custom config
            specialized_feature_cls = feature_cls.specialize(
                model_name, **feature_attrs
            )
            feature_classes[feature_cls.identifier].append(specialized_feature_cls)

        for attr in visited_attrs:
            del attr_dict[attr]

        return feature_classes


class EndPointMap(object):
    """internal endpoint map for spectrometer classes"""

    def __init__(
        self, ep_out=None, lowspeed_in=None, highspeed_in=None, highspeed_in2=None
    ):
        self.primary_out = self.ep_out = ep_out
        self.primary_in = self.lowspeed_in = lowspeed_in
        self.secondary_out = ep_out
        self.secondary_in = self.highspeed_in = highspeed_in
        self.secondary_in2 = self.highspeed_in2 = highspeed_in2


class DarkPixelIndices(tuple):
    """internal dark pixel range class"""

    def __new__(cls, indices):
        """dark pixel indices

        Parameters
        ----------
        indices : iterable
            index of electric dark pixel
        """
        return super(DarkPixelIndices, cls).__new__(
            DarkPixelIndices, sorted(set(indices))
        )

    @classmethod
    def from_ranges(cls, *ranges):
        """return dark pixes indices from ranges

        Parameters
        ----------
        *ranges : (`int`, `int`)
            ranges of electric dark pixels
        """
        dp = itertools.chain(*(range(low, high) for (low, high) in ranges))
        # noinspection PyArgumentList
        return cls(dp)


class TriggerMode(enum.IntEnum):
    """internal trigger modes enum"""

    NORMAL = 0x00
    SOFTWARE = 0x01
    LEVEL = 0x01
    SYNCHRONIZATION = 0x02
    HARDWARE = 0x03
    EDGE = 0x03
    SINGLE_SHOT = 0x04
    SELF_NORMAL = 0x80
    SELF_SOFTWARE = 0x81
    SELF_SYNCHRONIZATION = 0x82
    SELF_HARDWARE = 0x83
    DISABLED = 0xFF
    OBP_NORMAL = 0x00
    OBP_EXTERNAL = 0x01
    OBP_INTERNAL = 0x02
    OBP_EDGE = 0x01
    OBP_LEVEL = 0x03

    @classmethod
    def supported(cls, *mode_strings):
        return set(getattr(cls, mode_string) for mode_string in mode_strings)


class SeaBreezeDevice(with_metaclass(_SeaBreezeDeviceMeta)):

    # internal attribute
    _model_name = None
    _serial_number = "?"
    _cached_features = None

    def __new__(cls, raw_device=None):
        if raw_device is None:
            raise SeaBreezeError(
                "Don't instantiate SeaBreezeDevice directly. Use `SeabreezeAPI.list_devices()`."
            )
        for transport in {USBTransport}:
            supported_model = transport.supported_model(raw_device)
            if supported_model is not None:
                break
        else:
            raise TypeError("No transport supports device.")
        specialized_cls = _model_class_registry[supported_model]
        return super(SeaBreezeDevice, cls).__new__(specialized_cls)

    def __init__(self, raw_device=None):
        if raw_device is None:
            raise SeaBreezeError(
                "Don't instantiate SeaBreezeDevice directly. Use `SeabreezeAPI.list_devices()`."
            )
        self._raw_device = raw_device
        for transport in self._transport_classes:
            if transport.supported_model(self._raw_device) is not None:
                self._transport = transport()
                break
        else:
            raise TypeError("No transport supports device.")
        try:
            self._serial_number = self.get_serial_number()
        except SeaBreezeError:
            pass

    @property
    def model(self):
        return self._model_name

    @property
    def serial_number(self):
        return self._serial_number

    def __repr__(self):
        return "<SeaBreezeDevice %s:%s>" % (self.model, self.serial_number)

    def open(self):
        """open the spectrometer usb connection

        Returns
        -------
        None
        """
        self._transport.open_device(self._raw_device)
        # cache features
        self._cached_features = {}
        _ = self.features
        # get serial
        self._serial_number = self.get_serial_number()

    def close(self):
        """close the spectrometer usb connection

        Returns
        -------
        None
        """
        if self.is_open:
            self._transport.close_device()

    @property
    def is_open(self):
        """returns if the spectrometer device usb connection is opened

        Returns
        -------
        bool
        """
        return self._transport.is_open

    def get_serial_number(self):
        """return the serial number string of the spectrometer

        Returns
        -------
        serial_number: str
        """
        try:
            protocol = self._transport.protocol
            if protocol is None:
                raise AttributeError("transport not opened")

            elif isinstance(protocol, OOIProtocol):
                # The serial is stored in slot 0
                # noinspection PyUnresolvedReferences
                return self.f.eeprom.eeprom_read_slot(0)

            elif isinstance(protocol, OBPProtocol):
                serial_len = ord(protocol.query(0x00000101))
                return protocol.query(0x00000100)[:serial_len].decode("utf8")

            else:
                raise NotImplementedError(
                    "No serial number for protocol class {}".format(
                        protocol.__class__.__name__
                    )
                )
        except AttributeError:
            raise SeaBreezeError("device not open")

    @property
    def features(self):
        """return a dictionary of all supported features

        this returns a dictionary with all supported Features of the spectrometer

        Returns
        -------
        features : `dict` [`str`, `seabreeze.cseabreeze.SeaBreezeFeature`]
        """
        if not self._cached_features:
            protocol = self._transport.protocol
            self._cached_features = {}
            for identifier in sbf.SeaBreezeFeature.get_feature_class_registry():
                f_list = self._cached_features.setdefault(identifier, [])
                for feature_cls in self._feature_classes[identifier]:
                    assert (
                        issubclass(feature_cls, sbf.SeaBreezeFeature)
                        and identifier == feature_cls.identifier
                    )
                    if not feature_cls.supports_protocol(protocol):
                        continue
                    f_list.append(feature_cls(self._transport.protocol, len(f_list)))
        return self._cached_features

    @property
    def f(self):
        """convenience access to features via attributes

        this allows you to access a feature like this::

            # via .features
            device.features['spectrometer'][0].get_intensities()
            # via .f
            device.f.spectrometer.get_intensities()

        """

        class FeatureAccessHandler(object):
            def __init__(self, feature_dict):
                for identifier, features in feature_dict.items():
                    setattr(
                        self, identifier, features[0] if features else None
                    )  # TODO: raise FeatureNotAvailable?

        return FeatureAccessHandler(self.features)


# SPECTROMETER DEFINITIONS
# ========================
#
class USB2000PLUS(SeaBreezeDevice):

    model_name = "USB2000PLUS"

    # communication config
    transport = (USBTransport,)
    usb_product_id = 0x101E
    usb_endpoint_map = EndPointMap(
        ep_out=0x01, lowspeed_in=0x81, highspeed_in=0x82, highspeed_in2=0x86
    )
    usb_protocol = OOIProtocol

    # spectrometer config
    dark_pixel_indices = DarkPixelIndices.from_ranges((6, 21))  # as in seabreeze-3.0.9
    integration_time_min = 1000
    integration_time_max = 655350000
    integration_time_base = 1
    spectrum_num_pixel = 2048
    spectrum_raw_length = (2048 * 2) + 1
    spectrum_max_value = 65535
    trigger_modes = TriggerMode.supported(
        "NORMAL", "SOFTWARE", "SYNCHRONIZATION", "HARDWARE"
    )

    # features
    feature_classes = (
        sbf.eeprom.SeaBreezeEEPromFeatureOOI,
        sbf.spectrometer.SeaBreezeSpectrometerFeatureUSB2000PLUS,
        sbf.continuousstrobe.SeaBreezeContinuousStrobeFeatureOOI,
        sbf.rawusb.SeaBreezeRawUSBBusAccessFeature,
    )


class USB2000(SeaBreezeDevice):

    model_name = "USB2000"

    # communication config
    transport = (USBTransport,)
    usb_product_id = 0x1002
    usb_endpoint_map = EndPointMap(ep_out=0x02, lowspeed_in=0x87, highspeed_in=0x82)
    usb_protocol = OOIProtocol

    # spectrometer config
    dark_pixel_indices = DarkPixelIndices.from_ranges((2, 24))
    integration_time_min = 3000
    integration_time_max = 655350000
    integration_time_base = 1000
    spectrum_num_pixel = 2048
    spectrum_raw_length = (2048 * 2) + 1
    spectrum_max_value = 4095
    trigger_modes = TriggerMode.supported("NORMAL", "SOFTWARE", "HARDWARE")

    # features
    feature_classes = (
        sbf.eeprom.SeaBreezeEEPromFeatureOOI,
        sbf.spectrometer.SeaBreezeSpectrometerFeatureUSB2000,
        sbf.rawusb.SeaBreezeRawUSBBusAccessFeature,
    )


class HR2000(SeaBreezeDevice):

    model_name = "HR2000"

    # communication config
    transport = (USBTransport,)
    usb_product_id = 0x100A
    usb_endpoint_map = EndPointMap(ep_out=0x02, lowspeed_in=0x87, highspeed_in=0x82)
    usb_protocol = OOIProtocol

    # spectrometer config
    dark_pixel_indices = DarkPixelIndices.from_ranges((2, 24))
    integration_time_min = 3000
    integration_time_max = 655350000
    integration_time_base = 1000
    spectrum_num_pixel = 2048
    spectrum_raw_length = (2048 * 2) + 1
    spectrum_max_value = 4095
    trigger_modes = TriggerMode.supported("NORMAL", "SOFTWARE", "HARDWARE")

    # features
    feature_classes = (
        sbf.eeprom.SeaBreezeEEPromFeatureOOI,
        sbf.spectrometer.SeaBreezeSpectrometerFeatureHR2000,
        sbf.rawusb.SeaBreezeRawUSBBusAccessFeature,
    )


class HR4000(SeaBreezeDevice):

    model_name = "HR4000"

    # communication config
    transport = (USBTransport,)
    usb_product_id = 0x1012
    usb_endpoint_map = EndPointMap(
        ep_out=0x01, lowspeed_in=0x81, highspeed_in=0x82, highspeed_in2=0x86
    )
    usb_protocol = OOIProtocol

    # spectrometer config
    dark_pixel_indices = DarkPixelIndices.from_ranges((2, 13))
    integration_time_min = 10
    integration_time_max = 655350000
    integration_time_base = 1
    spectrum_num_pixel = 3840
    spectrum_raw_length = (3840 * 2) + 1
    spectrum_max_value = 16383
    trigger_modes = TriggerMode.supported(
        "NORMAL", "SOFTWARE", "SYNCHRONIZATION", "HARDWARE"
    )

    # features
    feature_classes = (
        sbf.eeprom.SeaBreezeEEPromFeatureOOI,
        sbf.spectrometer.SeaBreezeSpectrometerFeatureHR4000,
        sbf.rawusb.SeaBreezeRawUSBBusAccessFeature,
    )


class HR2000PLUS(SeaBreezeDevice):

    model_name = "HR2000PLUS"

    # communication config
    transport = (USBTransport,)
    usb_product_id = 0x1016
    usb_endpoint_map = EndPointMap(
        ep_out=0x01, lowspeed_in=0x81, highspeed_in=0x82, highspeed_in2=0x86
    )
    usb_protocol = OOIProtocol

    # spectrometer config
    dark_pixel_indices = DarkPixelIndices.from_ranges((2, 24))
    integration_time_min = 1000
    integration_time_max = 655350000
    integration_time_base = 1
    spectrum_num_pixel = 2048
    spectrum_raw_length = (2048 * 2) + 1
    spectrum_max_value = 16383
    trigger_modes = TriggerMode.supported(
        "NORMAL", "SOFTWARE", "SYNCHRONIZATION", "HARDWARE"
    )

    # features
    feature_classes = (
        sbf.eeprom.SeaBreezeEEPromFeatureOOI,
        sbf.spectrometer.SeaBreezeSpectrometerFeatureHR2000PLUS,
        sbf.rawusb.SeaBreezeRawUSBBusAccessFeature,
    )


class USB650(SeaBreezeDevice):

    model_name = "USB650"

    # communication config
    transport = (USBTransport,)
    usb_product_id = 0x1014
    usb_endpoint_map = EndPointMap(ep_out=0x02, lowspeed_in=0x87, highspeed_in=0x82)
    usb_protocol = OOIProtocol

    # spectrometer config
    dark_pixel_indices = DarkPixelIndices.from_ranges()
    integration_time_min = 3000
    integration_time_max = 655350000
    integration_time_base = 1000
    spectrum_num_pixel = 2048
    spectrum_raw_length = (2048 * 2) + 1
    spectrum_max_value = 4095
    trigger_modes = TriggerMode.supported("NORMAL", "SOFTWARE", "HARDWARE")

    # features
    feature_classes = (
        sbf.eeprom.SeaBreezeEEPromFeatureOOI,
        sbf.spectrometer.SeaBreezeSpectrometerFeatureUSB650,
        sbf.rawusb.SeaBreezeRawUSBBusAccessFeature,
        sbf.continuousstrobe.SeaBreezeContinuousStrobeFeatureOOI,
    )


class QE65000(SeaBreezeDevice):

    model_name = "QE65000"

    # communication config
    transport = (USBTransport,)
    usb_product_id = 0x1018
    usb_endpoint_map = EndPointMap(
        ep_out=0x01, lowspeed_in=0x81, highspeed_in=0x82, highspeed_in2=0x86
    )
    usb_protocol = OOIProtocol

    # spectrometer config
    dark_pixel_indices = DarkPixelIndices.from_ranges(
        (0, 4), (1040, 1044)
    )  # as in seabreeze-3.0.5
    integration_time_min = 8000
    integration_time_max = 1600000000
    integration_time_base = 1000
    spectrum_num_pixel = 1280
    spectrum_raw_length = (1024 + 256) * 2 + 1
    spectrum_max_value = 65535
    trigger_modes = TriggerMode.supported("NORMAL", "SOFTWARE", "HARDWARE")

    # features
    feature_classes = (
        sbf.eeprom.SeaBreezeEEPromFeatureOOI,
        sbf.spectrometer.SeaBreezeSpectrometerFeatureQE65000,
        sbf.rawusb.SeaBreezeRawUSBBusAccessFeature,
    )


class USB4000(SeaBreezeDevice):

    model_name = "USB4000"

    # communication config
    transport = (USBTransport,)
    usb_product_id = 0x1022
    usb_endpoint_map = EndPointMap(
        ep_out=0x01, lowspeed_in=0x81, highspeed_in=0x82, highspeed_in2=0x86
    )
    usb_protocol = OOIProtocol

    # spectrometer config
    dark_pixel_indices = (
        DarkPixelIndices.from_ranges((5, 16)),
    )  # as in seabreeze-3.0.9
    integration_time_min = 10
    integration_time_max = 65535000
    integration_time_base = 1
    spectrum_num_pixel = 3840
    spectrum_raw_length = (3840 * 2) + 1
    spectrum_max_value = 65535
    trigger_modes = TriggerMode.supported(
        "NORMAL", "SOFTWARE", "SYNCHRONIZATION", "HARDWARE"
    )

    # features
    feature_classes = (
        sbf.eeprom.SeaBreezeEEPromFeatureOOI,
        sbf.spectrometer.SeaBreezeSpectrometerFeatureUSB4000,
        sbf.rawusb.SeaBreezeRawUSBBusAccessFeature,
    )


class NIRQUEST512(SeaBreezeDevice):

    model_name = "NIRQUEST512"

    # communication config
    transport = (USBTransport,)
    usb_product_id = 0x1026
    usb_endpoint_map = EndPointMap(
        ep_out=0x01, lowspeed_in=0x81, highspeed_in=0x82, highspeed_in2=0x86
    )
    usb_protocol = OOIProtocol

    # spectrometer config
    dark_pixel_indices = (DarkPixelIndices.from_ranges(),)  # as in seabreeze-3.0.9
    integration_time_min = 1000
    integration_time_max = 1600000000
    integration_time_base = 1000
    spectrum_num_pixel = 512
    spectrum_raw_length = (512 * 2) + 1
    spectrum_max_value = 65535
    trigger_modes = TriggerMode.supported(
        "NORMAL", "SOFTWARE", "SYNCHRONIZATION", "HARDWARE"
    )

    # features
    feature_classes = (
        sbf.eeprom.SeaBreezeEEPromFeatureOOI,
        sbf.spectrometer.SeaBreezeSpectrometerFeatureNIRQUEST512,
        sbf.rawusb.SeaBreezeRawUSBBusAccessFeature,
    )


class NIRQUEST256(SeaBreezeDevice):

    model_name = "NIRQUEST256"

    # communication config
    transport = (USBTransport,)
    usb_product_id = 0x1028
    usb_endpoint_map = EndPointMap(
        ep_out=0x01, lowspeed_in=0x81, highspeed_in=0x82, highspeed_in2=0x86
    )
    usb_protocol = OOIProtocol

    # spectrometer config
    dark_pixel_indices = (DarkPixelIndices.from_ranges(),)  # as in seabreeze-3.0.9
    integration_time_min = 1000
    integration_time_max = 1600000000
    integration_time_base = 1000
    spectrum_num_pixel = 256
    spectrum_raw_length = (256 * 2) + 1
    spectrum_max_value = 65535
    trigger_modes = TriggerMode.supported(
        "NORMAL", "SOFTWARE", "SYNCHRONIZATION", "HARDWARE"
    )

    # features
    feature_classes = (
        sbf.eeprom.SeaBreezeEEPromFeatureOOI,
        sbf.spectrometer.SeaBreezeSpectrometerFeatureNIRQUEST256,
        sbf.rawusb.SeaBreezeRawUSBBusAccessFeature,
    )


class MAYA2000PRO(SeaBreezeDevice):

    model_name = "MAYA2000PRO"

    # communication config
    transport = (USBTransport,)
    usb_product_id = 0x102A
    usb_endpoint_map = EndPointMap(
        ep_out=0x01, lowspeed_in=0x81, highspeed_in=0x82, highspeed_in2=0x86
    )
    usb_protocol = OOIProtocol

    # spectrometer config
    dark_pixel_indices = DarkPixelIndices.from_ranges((0, 4), (2064, 2068))
    integration_time_min = 7200
    integration_time_max = 65000000
    integration_time_base = 1
    spectrum_num_pixel = 2304
    spectrum_raw_length = (2304 * 2) + 1
    spectrum_max_value = 64000
    trigger_modes = TriggerMode.supported(
        "NORMAL", "SOFTWARE", "SYNCHRONIZATION", "HARDWARE"
    )

    # features
    feature_classes = (
        sbf.eeprom.SeaBreezeEEPromFeatureOOI,
        sbf.spectrometer.SeaBreezeSpectrometerFeatureMAYA2000PRO,
        sbf.rawusb.SeaBreezeRawUSBBusAccessFeature,
    )


class MAYA2000(SeaBreezeDevice):

    model_name = "MAYA2000"

    # communication config
    transport = (USBTransport,)
    usb_product_id = 0x102C
    usb_endpoint_map = EndPointMap(
        ep_out=0x01, lowspeed_in=0x81, highspeed_in=0x82, highspeed_in2=0x86
    )
    usb_protocol = OOIProtocol

    # spectrometer config
    dark_pixel_indices = DarkPixelIndices.from_ranges((0, 8), (2072, 2080))
    integration_time_min = 15000
    integration_time_max = 1600000000
    integration_time_base = 1
    spectrum_num_pixel = 2304
    spectrum_raw_length = (2304 * 2) + 1
    spectrum_max_value = 65535
    trigger_modes = TriggerMode.supported("NORMAL", "SOFTWARE", "HARDWARE")

    # features
    feature_classes = (
        sbf.eeprom.SeaBreezeEEPromFeatureOOI,
        sbf.spectrometer.SeaBreezeSpectrometerFeatureMAYA2000,
        sbf.rawusb.SeaBreezeRawUSBBusAccessFeature,
    )


class TORUS(SeaBreezeDevice):

    model_name = "TORUS"

    # communication config
    transport = (USBTransport,)
    usb_product_id = 0x1040
    usb_endpoint_map = EndPointMap(
        ep_out=0x01, lowspeed_in=0x81, highspeed_in=0x82, highspeed_in2=0x86
    )
    usb_protocol = OOIProtocol

    # spectrometer config
    dark_pixel_indices = DarkPixelIndices.from_ranges()
    integration_time_min = 1000
    integration_time_max = 655350000
    integration_time_base = 1
    spectrum_num_pixel = 2048
    spectrum_raw_length = (2048 * 2) + 1
    spectrum_max_value = 65535
    trigger_modes = TriggerMode.supported(
        "NORMAL", "SOFTWARE", "SYNCHRONIZATION", "HARDWARE"
    )

    # features
    feature_classes = (
        sbf.eeprom.SeaBreezeEEPromFeatureOOI,
        sbf.spectrometer.SeaBreezeSpectrometerFeatureTORUS,
        sbf.rawusb.SeaBreezeRawUSBBusAccessFeature,
    )


class APEX(SeaBreezeDevice):

    model_name = "APEX"

    # communication config
    transport = (USBTransport,)
    usb_product_id = 0x1044
    usb_endpoint_map = EndPointMap(
        ep_out=0x01, lowspeed_in=0x81, highspeed_in=0x82, highspeed_in2=0x86
    )
    usb_protocol = OOIProtocol

    # spectrometer config
    dark_pixel_indices = DarkPixelIndices.from_ranges((0, 4), (2064, 2068))
    integration_time_min = 15000
    integration_time_max = 1600000000
    integration_time_base = 1
    spectrum_num_pixel = 2304
    spectrum_raw_length = (2304 * 2) + 1
    spectrum_max_value = 64000
    trigger_modes = TriggerMode.supported("NORMAL",)

    # features
    feature_classes = (
        sbf.eeprom.SeaBreezeEEPromFeatureOOI,
        sbf.spectrometer.SeaBreezeSpectrometerFeatureAPEX,
        sbf.rawusb.SeaBreezeRawUSBBusAccessFeature,
    )


class MAYALSL(SeaBreezeDevice):

    model_name = "MAYALSL"

    # communication config
    transport = (USBTransport,)
    usb_product_id = 0x1046
    usb_endpoint_map = EndPointMap(
        ep_out=0x01, lowspeed_in=0x81, highspeed_in=0x82, highspeed_in2=0x86
    )
    usb_protocol = OOIProtocol

    # spectrometer config
    dark_pixel_indices = DarkPixelIndices.from_ranges((0, 4), (2064, 2068))
    integration_time_min = 7200
    integration_time_max = 65000000
    integration_time_base = 1
    spectrum_num_pixel = 2304
    spectrum_raw_length = (2304 * 2) + 1
    spectrum_max_value = 64000
    trigger_modes = TriggerMode.supported(
        "NORMAL", "SOFTWARE", "SYNCHRONIZATION", "HARDWARE"
    )

    # features
    feature_classes = (
        sbf.eeprom.SeaBreezeEEPromFeatureOOI,
        sbf.spectrometer.SeaBreezeSpectrometerFeatureMAYALSL,
        sbf.rawusb.SeaBreezeRawUSBBusAccessFeature,
    )


class JAZ(SeaBreezeDevice):

    model_name = "JAZ"

    # communication config
    transport = (USBTransport,)
    usb_product_id = 0x2000
    usb_endpoint_map = EndPointMap(ep_out=0x01, lowspeed_in=0x81, highspeed_in=0x82)
    usb_protocol = OOIProtocol

    # spectrometer config
    dark_pixel_indices = DarkPixelIndices.from_ranges((2, 24))
    integration_time_min = 1000
    integration_time_max = 655350000
    integration_time_base = 1
    spectrum_num_pixel = 2048
    spectrum_raw_length = 2048 * 2  # XXX: No Sync byte!
    spectrum_max_value = 65535
    trigger_modes = TriggerMode.supported(
        "NORMAL", "SOFTWARE", "SYNCHRONIZATION", "HARDWARE"
    )

    # features
    feature_classes = (
        sbf.eeprom.SeaBreezeEEPromFeatureOOI,
        sbf.spectrometer.SeaBreezeSpectrometerFeatureJAZ,
        sbf.rawusb.SeaBreezeRawUSBBusAccessFeature,
    )


class STS(SeaBreezeDevice):

    model_name = "STS"

    # communication config
    transport = (USBTransport,)
    usb_product_id = 0x4000
    usb_endpoint_map = EndPointMap(
        ep_out=0x01, lowspeed_in=0x81
    )  # XXX: we'll ignore the alternative EPs
    usb_protocol = OBPProtocol

    # spectrometer config
    dark_pixel_indices = DarkPixelIndices.from_ranges()
    integration_time_min = 10
    integration_time_max = 85000000
    integration_time_base = 1
    spectrum_num_pixel = 1024
    spectrum_raw_length = 1024 * 2
    spectrum_max_value = 16383
    trigger_modes = TriggerMode.supported("OBP_NORMAL", "OBP_EXTERNAL", "OBP_INTERNAL")

    # features
    feature_classes = (
        sbf.spectrometer.SeaBreezeSpectrometerFeatureSTS,
        sbf.rawusb.SeaBreezeRawUSBBusAccessFeature,
    )


class QEPRO(SeaBreezeDevice):

    model_name = "QEPRO"

    # communication config
    transport = (USBTransport,)
    usb_product_id = 0x4004
    usb_endpoint_map = EndPointMap(
        ep_out=0x01, lowspeed_in=0x81
    )  # XXX: we'll ignore the alternative EPs
    usb_protocol = OBPProtocol

    # spectrometer config
    dark_pixel_indices = DarkPixelIndices.from_ranges((0, 4), (1040, 1044))
    integration_time_min = 10000
    integration_time_max = 1600000000
    integration_time_base = 1
    spectrum_num_pixel = 1044
    spectrum_raw_length = (1044 * 4) + 32  # XXX: Metadata
    spectrum_max_value = (2 ** 18) - 1
    trigger_modes = TriggerMode.supported("NORMAL", "LEVEL", "SYNCHRONIZATION", "EDGE")

    # features
    feature_classes = (
        sbf.spectrometer.SeaBreezeSpectrometerFeatureQEPRO,
        sbf.rawusb.SeaBreezeRawUSBBusAccessFeature,
    )


class VENTANA(SeaBreezeDevice):

    model_name = "VENTANA"

    # communication config
    transport = (USBTransport,)
    usb_product_id = 0x5000
    usb_endpoint_map = EndPointMap(ep_out=0x01, lowspeed_in=0x82)
    usb_protocol = OBPProtocol

    # spectrometer config
    dark_pixel_indices = DarkPixelIndices.from_ranges()
    integration_time_min = 22000
    integration_time_max = 60000000
    integration_time_base = 1
    spectrum_num_pixel = 1024
    spectrum_raw_length = 1024 * 2  # XXX: No Sync byte!
    spectrum_max_value = 65535
    trigger_modes = TriggerMode.supported("NORMAL", "LEVEL", "SYNCHRONIZATION", "EDGE")

    # features
    feature_classes = (
        sbf.spectrometer.SeaBreezeSpectrometerFeatureVENTANA,
        sbf.rawusb.SeaBreezeRawUSBBusAccessFeature,
    )


class SPARK(SeaBreezeDevice):

    model_name = "SPARK"

    # communication config
    transport = (USBTransport,)
    usb_product_id = 0x4200
    usb_endpoint_map = EndPointMap(
        ep_out=0x01, lowspeed_in=0x81
    )  # XXX: we'll ignore the alternative EPs
    usb_protocol = OBPProtocol

    # spectrometer config
    dark_pixel_indices = DarkPixelIndices.from_ranges()
    integration_time_min = 10
    integration_time_max = 85000000
    integration_time_base = 1
    spectrum_num_pixel = 1024
    spectrum_raw_length = (1024 * 2) + 64  # XXX: Metadata
    spectrum_max_value = 16383
    trigger_modes = TriggerMode.supported("OBP_NORMAL", "OBP_EXTERNAL", "OBP_INTERNAL")

    # features
    feature_classes = (
        sbf.spectrometer.SeaBreezeSpectrometerFeatureSPARK,
        sbf.rawusb.SeaBreezeRawUSBBusAccessFeature,
    )


class HDX(SeaBreezeDevice):

    model_name = "HDX"

    # communication config
    transport = (USBTransport,)
    usb_product_id = 0x2003
    usb_endpoint_map = EndPointMap(
        ep_out=0x01, lowspeed_in=0x81, highspeed_in=0x82, highspeed_in2=0x86
    )
    usb_protocol = OBPProtocol

    # spectrometer config
    dark_pixel_indices = DarkPixelIndices.from_ranges()
    integration_time_min = 6000
    integration_time_max = 10000000
    integration_time_base = 1
    spectrum_num_pixel = 2068
    spectrum_raw_length = (2068 * 2) + 64  # XXX: Metadata
    spectrum_max_value = 65535
    trigger_modes = TriggerMode.supported(
        "OBP_NORMAL", "OBP_LEVEL", "OBP_EDGE", "DISABLED"
    )

    # features
    feature_classes = (
        sbf.spectrometer.SeaBreezeSpectrometerFeatureHDX,
        sbf.rawusb.SeaBreezeRawUSBBusAccessFeature,
        sbf.nonlinearity.NonlinearityCoefficientsFeatureOBP,
    )



import math
import struct
import time

from seabreeze.pyseabreeze.features._base import SeaBreezeFeature
from seabreeze.pyseabreeze.protocol import OOIProtocol


# Definition
# ==========
class SeaBreezeContinuousStrobeFeature(SeaBreezeFeature):
    identifier = "continuous_strobe"

    def set_enable(self, strobe_enable):
        raise NotImplementedError("implement in derived class")

    def set_period_micros(self, period_micros):
        raise NotImplementedError("implement in derived class")


class _FPGARegisterFeatureOOI(object):
    class Codes(object):
        FIRMWARE_VERSION = 0x04
        V1_CONTINUOUS_STROBE_TIMER_INTERVAL_DIVISOR = (
            V3_CONTINUOUS_STROBE_TIMER_MSB
        ) = 0x08
        V1_CONTINUOUS_STROBE_BASE_CLOCK_DIVISOR = V3_CONTINUOUS_STROBE_TIMER_LSB = 0x0C

    def __init__(self, protocol):
        self.protocol = protocol

    def get_major_version(self):
        fw_raw = self.protocol.query(0x6B, self.Codes.FIRMWARE_VERSION)
        command, data = struct.unpack("<BH", fw_raw)
        assert command == self.Codes.FIRMWARE_VERSION
        fw_version = (
            (data >> 12) & 0x0F,  # major
            (data >> 4) & 0xFF,  # minor
            data & 0xF,  # patch
        )
        return fw_version[0]

    def write_register(self, register, data):
        self.protocol.send(0x6A, payload=(register, data))
        time.sleep(0.0001)  # guarantee 100us sleep between commands


class SeaBreezeContinuousStrobeFeatureOOI(SeaBreezeContinuousStrobeFeature):
    _required_protocol_cls = OOIProtocol

    def __init__(self, protocol, feature_id, **kwargs):
        super(SeaBreezeContinuousStrobeFeatureOOI, self).__init__(
            protocol, feature_id, **kwargs
        )
        self._fpga = _FPGARegisterFeatureOOI(protocol)

    def set_enable(self, strobe_enable):
        """
        Sets the Lamp Enable line (J2 pin 4) as follows.
        The Single Strobe and Continuous Strobe signals are enabled/disabled by this Lamp Enable Signal.

        Parameters
        ----------
        strobe_enable: `bool`
            False: Lamp Enable Low/Off
            True: Lamp Enable HIGH/On

        Returns
        -------
        None
        """
        self.protocol.send(0x03, int(strobe_enable))

    def set_period_micros(self, period_micros):
        """set continuous strobe period in microseconds

        Parameters
        ----------
        period_micros : `int`
            period in microseconds 0 < period_micros <~ 60 seconds

        Returns
        -------
        None
        """
        period_micros = int(period_micros)
        if period_micros <= 0:
            raise ValueError("requires period_micros > 0")

        fpga_major_version = self._fpga.get_major_version()

        # ported from cseabreeze
        if fpga_major_version == 1:
            # The base clock value is 48Mhz, so divide out the 48 leaving a 1 usec resolution
            # Compute how many bits are needed to represent the entire amount.
            # The first 10 will be absorbed by timerValue, and up to 16 more by baseClockValue.
            # If more than 26 bits (64 seconds) are given, it is too large.
            bits = int(math.ceil(math.log(period_micros, 2)))

            if bits <= 16:  # 0-~1023 usec
                timer_interval = 48  # use 5.6 bits (leaving about 10)
                base_clock = period_micros - 1

            elif bits <= 26:  # up to about 64 seconds
                timer_interval = 48000  # = 2^(15.55), about 1ms (new minimum step size)
                base_clock = int(period_micros / 1000) - 1

            else:
                raise ValueError("period_micros is too large")

            self._fpga.write_register(
                self._fpga.Codes.V1_CONTINUOUS_STROBE_TIMER_INTERVAL_DIVISOR,
                timer_interval,
            )
            self._fpga.write_register(
                self._fpga.Codes.V1_CONTINUOUS_STROBE_BASE_CLOCK_DIVISOR, base_clock
            )

        elif fpga_major_version == 3:
            counts = period_micros * 48
            if not 0 < counts < 2 ** 32:
                raise ValueError("period_micros is too large")
            self._fpga.write_register(
                self._fpga.Codes.V3_CONTINUOUS_STROBE_TIMER_MSB, (counts >> 16) & 0xFFFF
            )
            self._fpga.write_register(
                self._fpga.Codes.V3_CONTINUOUS_STROBE_TIMER_LSB, counts & 0xFFFF
            )

        else:
            raise RuntimeError("unsupported FPGA major version")



"""This is a wrapper for the pyusb-implementation of the seabreeze-library

"""
from seabreeze.pyseabreeze.api import SeaBreezeAPI
from seabreeze.pyseabreeze.devices import SeaBreezeDevice
from seabreeze.pyseabreeze.exceptions import SeaBreezeError, SeaBreezeNumFeaturesError

# noinspection PyUnresolvedReferences
from seabreeze.pyseabreeze.features import (
    SeaBreezeAcquisitionDelayFeature,
    SeaBreezeContinuousStrobeFeature,
    SeaBreezeDataBufferFeature,
    SeaBreezeDHCPServerFeature,
    SeaBreezeEEPROMFeature,
    SeaBreezeEthernetConfigurationFeature,
    SeaBreezeFastBufferFeature,
    SeaBreezeGPIOFeature,
    SeaBreezeI2CMasterFeature,
    SeaBreezeIntrospectionFeature,
    SeaBreezeIPv4Feature,
    SeaBreezeIrradCalFeature,
    SeaBreezeLightSourceFeature,
    SeaBreezeMulticastFeature,
    SeaBreezeNetworkConfigurationFeature,
    SeaBreezeNonlinearityCoefficientsFeature,
    SeaBreezeOpticalBenchFeature,
    SeaBreezePixelBinningFeature,
    SeaBreezeRawUSBBusAccessFeature,
    SeaBreezeRevisionFeature,
    SeaBreezeShutterFeature,
    SeaBreezeSpectrometerFeature,
    SeaBreezeSpectrumProcessingFeature,
    SeaBreezeStrayLightCoefficientsFeature,
    SeaBreezeStrobeLampFeature,
    SeaBreezeTemperatureFeature,
    SeaBreezeThermoElectricFeature,
    SeaBreezeWifiConfigurationFeature,
)

# noinspection PyUnresolvedReferences,PyProtectedMember
from seabreeze.pyseabreeze.features._base import SeaBreezeFeature

_backend_ = "pyseabreeze"



from seabreeze.pyseabreeze.features._base import SeaBreezeFeature


# Definition
# ==========
#
# TODO: This feature needs to be implemented for pyseabreeze
#
class SeaBreezeStrayLightCoefficientsFeature(SeaBreezeFeature):
    identifier = "stray_light_coefficients"

    def get_stray_light_coefficients(self):
        raise NotImplementedError("implement in derived class")



from seabreeze.pyseabreeze.features._base import SeaBreezeFeature


# Definition
# ==========
#
# TODO: This feature needs to be implemented for pyseabreeze
#
class SeaBreezeShutterFeature(SeaBreezeFeature):
    identifier = "shutter"

    def set_shutter_open(self, state):
        raise NotImplementedError("implement in derived class")
